@@ -25,8 +25,10 @@
 #include <string.h>
 #include <config.h>
 #include "gclue-mozilla.h"
+#include "gclue-3g-tower.h"
 #include "gclue-config.h"
 #include "gclue-error.h"
+#include "gclue-wifi.h"
 
 /**
  * SECTION:gclue-mozilla
@@ -40,6 +42,22 @@
  * its easy to switch to Google's API.
  **/
 
+struct _GClueMozillaPrivate
+{
+        GClueWifi *wifi;
+
+        GClue3GTower tower;
+        gboolean tower_valid;
+        gboolean tower_submitted;
+
+        gboolean bss_submitted;
+};
+
+G_DEFINE_TYPE_WITH_CODE (GClueMozilla,
+                         gclue_mozilla,
+                         G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (GClueMozilla))
+
 #define BSSID_LEN 6
 #define BSSID_STR_LEN 17
 #define MAX_SSID_LEN 32
@@ -102,12 +120,10 @@ get_bssid_from_bss (WPABSS *bss, char *bssid)
         return TRUE;
 }
 
-static const char *
-get_url (void)
+const char *
+gclue_mozilla_get_locate_url (GClueMozilla *mozilla)
 {
-        GClueConfig *config;
-
-        config = gclue_config_get_singleton ();
+        GClueConfig *config = gclue_config_get_singleton ();
 
         return gclue_config_get_wifi_url (config);
 }
@@ -136,25 +152,55 @@ error:
         return FALSE;
 }
 
+static gboolean
+towertec_to_radiotype (GClueTowerTec tec,
+                       const char **radiotype_p)
+{
+        switch (tec) {
+        case GCLUE_TOWER_TEC_2G:
+            *radiotype_p = "gsm";
+            break;
+        case GCLUE_TOWER_TEC_3G:
+            *radiotype_p = "wcdma";
+            break;
+        case GCLUE_TOWER_TEC_4G:
+            *radiotype_p = "lte";
+            break;
+        default:
+            *radiotype_p = NULL;
+            return FALSE;
+        }
+
+        return TRUE;
+}
+
 SoupMessage *
-gclue_mozilla_create_query (GList        *bss_list, /* As in Access Points */
-                            GClue3GTower *tower,
+gclue_mozilla_create_query (GClueMozilla  *mozilla,
+                            gboolean skip_tower,
+                            gboolean skip_bss,
+                            const char **query_data_description,
                             GError      **error)
 {
+        gboolean has_tower = FALSE, has_bss = FALSE;
         SoupMessage *ret = NULL;
         JsonBuilder *builder;
+        g_autoptr(GList) bss_list = NULL;
         JsonGenerator *generator;
         JsonNode *root_node;
         char *data;
         gsize data_len;
-        const char *uri;
+        const char *uri, *radiotype;
         guint n_non_ignored_bsss;
         GList *iter;
         gint64 mcc, mnc;
+        g_autoptr(GBytes) body = NULL;
 
         builder = json_builder_new ();
         json_builder_begin_object (builder);
 
+        if (mozilla->priv->wifi && !skip_bss) {
+                bss_list = gclue_wifi_get_bss_list (mozilla->priv->wifi);
+        }
         /* We send pure geoip query using empty object if both bss_list and
          * tower are NULL.
          *
@@ -172,11 +218,11 @@ gclue_mozilla_create_query (GList        *bss_list, /* As in Access Points */
                 n_non_ignored_bsss++;
         }
 
-        if (tower != NULL &&
-            operator_code_to_mcc_mnc (tower->opc, &mcc, &mnc)) {
-
+        if (mozilla->priv->tower_valid && !skip_tower &&
+            towertec_to_radiotype (mozilla->priv->tower.tec, &radiotype) &&
+            operator_code_to_mcc_mnc (mozilla->priv->tower.opc, &mcc, &mnc)) {
                 json_builder_set_member_name (builder, "radioType");
-                json_builder_add_string_value (builder, "gsm");
+                json_builder_add_string_value (builder, radiotype);
 
                 json_builder_set_member_name (builder, "cellTowers");
                 json_builder_begin_array (builder);
@@ -184,21 +230,21 @@ gclue_mozilla_create_query (GList        *bss_list, /* As in Access Points */
                 json_builder_begin_object (builder);
 
                 json_builder_set_member_name (builder, "cellId");
-                json_builder_add_int_value (builder, tower->cell_id);
+                json_builder_add_int_value (builder, mozilla->priv->tower.cell_id);
                 json_builder_set_member_name (builder, "mobileCountryCode");
                 json_builder_add_int_value (builder, mcc);
                 json_builder_set_member_name (builder, "mobileNetworkCode");
                 json_builder_add_int_value (builder, mnc);
                 json_builder_set_member_name (builder, "locationAreaCode");
-                json_builder_add_int_value (builder, tower->lac);
-                if (tower->tec == GCLUE_TOWER_TEC_4G) {
-                        json_builder_set_member_name (builder, "radioType");
-                        json_builder_add_string_value (builder, "lte");
-		}
+                json_builder_add_int_value (builder, mozilla->priv->tower.lac);
+                json_builder_set_member_name (builder, "radioType");
+                json_builder_add_string_value (builder, radiotype);
 
                 json_builder_end_object (builder);
 
                 json_builder_end_array (builder);
+
+                has_tower = TRUE;
         }
 
         if (n_non_ignored_bsss >= 2) {
@@ -222,6 +268,8 @@ gclue_mozilla_create_query (GList        *bss_list, /* As in Access Points */
                         strength_dbm = wpa_bss_get_signal (bss);
                         json_builder_add_int_value (builder, strength_dbm);
                         json_builder_end_object (builder);
+
+                        has_bss = TRUE;
                 }
                 json_builder_end_array (builder);
         }
@@ -236,15 +284,24 @@ gclue_mozilla_create_query (GList        *bss_list, /* As in Access Points */
         g_object_unref (builder);
         g_object_unref (generator);
 
-        uri = get_url ();
+        uri = gclue_mozilla_get_locate_url (mozilla);
         ret = soup_message_new ("POST", uri);
-        soup_message_set_request (ret,
-                                  "application/json",
-                                  SOUP_MEMORY_TAKE,
-                                  data,
-                                  data_len);
+        body = g_bytes_new_take (data, data_len);
+        soup_message_set_request_body_from_bytes (ret, "application/json", body);
         g_debug ("Sending following request to '%s':\n%s", uri, data);
 
+        if (query_data_description) {
+                if (has_tower && has_bss) {
+                        *query_data_description = "3GPP + WiFi";
+                } else if (has_tower) {
+                        *query_data_description = "3GPP";
+                } else if (has_bss) {
+                        *query_data_description = "WiFi";
+                } else {
+                        *query_data_description = "GeoIP";
+                }
+        }
+
         return ret;
 }
 
@@ -252,28 +309,32 @@ static gboolean
 parse_server_error (JsonObject *object, GError **error)
 {
         JsonObject *error_obj;
-        int code;
         const char *message;
 
         if (!json_object_has_member (object, "error"))
             return FALSE;
 
         error_obj = json_object_get_object_member (object, "error");
-        code = json_object_get_int_member (error_obj, "code");
-        message = json_object_get_string_member (error_obj, "message");
+        if (json_object_has_member (error_obj, "message")) {
+                message = json_object_get_string_member (error_obj, "message");
+        } else {
+                message = "Unknown error";
+        }
 
-        g_set_error_literal (error, G_IO_ERROR, code, message);
+        g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_FAILED, message);
 
         return TRUE;
 }
 
 GClueLocation *
 gclue_mozilla_parse_response (const char *json,
+                              const char *location_description,
                               GError    **error)
 {
-        JsonParser *parser;
+        g_autoptr(JsonParser) parser = NULL;
         JsonNode *node;
         JsonObject *object, *loc_object;
+        g_autofree char *desc_new = NULL;
         GClueLocation *location;
         gdouble latitude, longitude, accuracy;
 
@@ -288,54 +349,78 @@ gclue_mozilla_parse_response (const char *json,
         if (parse_server_error (object, error))
                 return NULL;
 
+        if (json_object_has_member (object, "fallback")) {
+                const char *fallback;
+
+                fallback = json_object_get_string_member (object, "fallback");
+                if (fallback && strlen (fallback)) {
+                        desc_new = g_strdup_printf ("%s fallback (from %s data)",
+                                                    fallback, location_description);
+                        location_description = desc_new;
+                }
+        }
+
         loc_object = json_object_get_object_member (object, "location");
         latitude = json_object_get_double_member (loc_object, "lat");
         longitude = json_object_get_double_member (loc_object, "lng");
 
         accuracy = json_object_get_double_member (object, "accuracy");
 
-        location = gclue_location_new (latitude, longitude, accuracy);
-
-        g_object_unref (parser);
+        location = gclue_location_new (latitude, longitude, accuracy,
+                                       location_description);
 
         return location;
 }
 
-static const char *
-get_submit_config (const char **nick)
+const char *
+gclue_mozilla_get_submit_url (GClueMozilla *mozilla)
 {
-        GClueConfig *config;
+        GClueConfig *config = gclue_config_get_singleton ();
 
-        config = gclue_config_get_singleton ();
-        if (!gclue_config_get_wifi_submit_data (config))
+        if (gclue_config_get_wifi_submit_data (config))
+                return gclue_config_get_wifi_submit_url (config);
+        else
                 return NULL;
-
-        *nick = gclue_config_get_wifi_submit_nick (config);
-
-        return gclue_config_get_wifi_submit_url (config);
 }
 
 SoupMessage *
-gclue_mozilla_create_submit_query (GClueLocation   *location,
-                                   GList           *bss_list, /* As in Access Points */
-                                   GClue3GTower    *tower,
+gclue_mozilla_create_submit_query (GClueMozilla  *mozilla,
+                                   GClueLocation   *location,
                                    GError         **error)
 {
         SoupMessage *ret = NULL;
+        SoupMessageHeaders *request_headers;
         JsonBuilder *builder;
         JsonGenerator *generator;
         JsonNode *root_node;
-        char *data, *timestr;
-        const char *url, *nick;
+        char *data;
+        g_autoptr(GList) bss_list = NULL;
+        const char *url, *nick, *radiotype;
         gsize data_len;
         GList *iter;
-        gdouble lat, lon, accuracy, altitude;
-        GDateTime *datetime;
+        gdouble lat, lon, accuracy, altitude, speed;
+        guint64 time_ms;
         gint64 mcc, mnc;
+        GClueConfig *config;
+        g_autoptr(GBytes) body = NULL;
 
-        url = get_submit_config (&nick);
+        if (mozilla->priv->bss_submitted &&
+            (!mozilla->priv->tower_valid ||
+             mozilla->priv->tower_submitted))
+        {
+                g_debug ("Already created submit req for this data (bss submitted %d; tower: valid %d submitted %d)",
+                         (int)mozilla->priv->bss_submitted,
+                         (int)mozilla->priv->tower_valid,
+                         (int)mozilla->priv->tower_submitted);
+                goto out;
+        }
+
+
+        url = gclue_mozilla_get_submit_url (mozilla);
         if (url == NULL)
                 goto out;
+        config = gclue_config_get_singleton ();
+        nick = gclue_config_get_wifi_submit_nick (config);
 
         builder = json_builder_new ();
         json_builder_begin_object (builder);
@@ -345,12 +430,19 @@ gclue_mozilla_create_submit_query (GClueLocation   *location,
 
         json_builder_begin_object (builder);
 
+        json_builder_set_member_name (builder, "timestamp");
+        time_ms = 1000 * gclue_location_get_timestamp (location);
+        json_builder_add_int_value (builder, time_ms);
+
+        json_builder_set_member_name (builder, "position");
+        json_builder_begin_object (builder);
+
         lat = gclue_location_get_latitude (location);
-        json_builder_set_member_name (builder, "lat");
+        json_builder_set_member_name (builder, "latitude");
         json_builder_add_double_value (builder, lat);
 
         lon = gclue_location_get_longitude (location);
-        json_builder_set_member_name (builder, "lon");
+        json_builder_set_member_name (builder, "longitude");
         json_builder_add_double_value (builder, lon);
 
         accuracy = gclue_location_get_accuracy (location);
@@ -365,22 +457,19 @@ gclue_mozilla_create_submit_query (GClueLocation   *location,
                 json_builder_add_double_value (builder, altitude);
         }
 
-        datetime = g_date_time_new_from_unix_local
-                (gclue_location_get_timestamp (location));
-        /* We need to be compatible with GLib 2.56 so we cannot use this:
-         * timestr = g_date_time_format_iso8601 (datetime);
-         * Construct the format manually instead: */
-        timestr = g_date_time_format (datetime, "%Y-%m-%dT%H:%M:%S%:::z");
-        json_builder_set_member_name (builder, "time");
-        json_builder_add_string_value (builder, timestr);
-        g_free (timestr);
-        g_date_time_unref (datetime);
+        speed = gclue_location_get_speed (location);
+        if (speed != GCLUE_LOCATION_SPEED_UNKNOWN) {
+                json_builder_set_member_name (builder, "speed");
+                json_builder_add_double_value (builder, speed);
+        }
 
-        json_builder_set_member_name (builder, "radioType");
-        json_builder_add_string_value (builder, "gsm");
+        json_builder_end_object (builder); /* position */
 
+        if (mozilla->priv->wifi) {
+                bss_list = gclue_wifi_get_bss_list (mozilla->priv->wifi);
+        }
         if (bss_list != NULL) {
-                json_builder_set_member_name (builder, "wifi");
+                json_builder_set_member_name (builder, "wifiAccessPoints");
                 json_builder_begin_array (builder);
 
                 for (iter = bss_list; iter != NULL; iter = iter->next) {
@@ -388,50 +477,57 @@ gclue_mozilla_create_submit_query (GClueLocation   *location,
                         char mac[BSSID_STR_LEN + 1] = { 0 };
                         gint16 strength_dbm;
                         guint16 frequency;
+                        guint age_ms;
 
                         if (gclue_mozilla_should_ignore_bss (bss))
                                 continue;
 
                         json_builder_begin_object (builder);
-                        json_builder_set_member_name (builder, "key");
+
+                        json_builder_set_member_name (builder, "macAddress");
                         get_bssid_from_bss (bss, mac);
                         json_builder_add_string_value (builder, mac);
 
-                        json_builder_set_member_name (builder, "signal");
+                        json_builder_set_member_name (builder, "signalStrength");
                         strength_dbm = wpa_bss_get_signal (bss);
                         json_builder_add_int_value (builder, strength_dbm);
 
                         json_builder_set_member_name (builder, "frequency");
                         frequency = wpa_bss_get_frequency (bss);
                         json_builder_add_int_value (builder, frequency);
+
+                        json_builder_set_member_name (builder, "age");
+                        age_ms = 1000 * wpa_bss_get_age (bss);
+                        json_builder_add_int_value (builder, age_ms);
+
                         json_builder_end_object (builder);
                 }
 
-                json_builder_end_array (builder); /* wifi */
+                json_builder_end_array (builder); /* wifiAccessPoints */
         }
 
-        if (tower != NULL &&
-            operator_code_to_mcc_mnc (tower->opc, &mcc, &mnc)) {
-
-                json_builder_set_member_name (builder, "cell");
+        if (mozilla->priv->tower_valid &&
+            towertec_to_radiotype (mozilla->priv->tower.tec, &radiotype) &&
+            operator_code_to_mcc_mnc (mozilla->priv->tower.opc, &mcc, &mnc)) {
+                json_builder_set_member_name (builder, "cellTowers");
                 json_builder_begin_array (builder);
 
                 json_builder_begin_object (builder);
 
-                json_builder_set_member_name (builder, "radio");
-                json_builder_add_string_value (builder, "gsm");
-                json_builder_set_member_name (builder, "cid");
-                json_builder_add_int_value (builder, tower->cell_id);
-                json_builder_set_member_name (builder, "mcc");
+                json_builder_set_member_name (builder, "radioType");
+                json_builder_add_string_value (builder, radiotype);
+                json_builder_set_member_name (builder, "cellId");
+                json_builder_add_int_value (builder, mozilla->priv->tower.cell_id);
+                json_builder_set_member_name (builder, "mobileCountryCode");
                 json_builder_add_int_value (builder, mcc);
-                json_builder_set_member_name (builder, "mnc");
+                json_builder_set_member_name (builder, "mobileNetworkCode");
                 json_builder_add_int_value (builder, mnc);
-                json_builder_set_member_name (builder, "lac");
-                json_builder_add_int_value (builder, tower->lac);
+                json_builder_set_member_name (builder, "locationAreaCode");
+                json_builder_add_int_value (builder, mozilla->priv->tower.lac);
 
                 json_builder_end_object (builder);
 
-                json_builder_end_array (builder); /* cell */
+                json_builder_end_array (builder); /* cellTowers */
         }
 
         json_builder_end_object (builder);
@@ -448,17 +544,18 @@ gclue_mozilla_create_submit_query (GClueLocation   *location,
         g_object_unref (generator);
 
         ret = soup_message_new ("POST", url);
+        request_headers = soup_message_get_request_headers (ret);
         if (nick != NULL && nick[0] != '\0')
-                soup_message_headers_append (ret->request_headers,
+                soup_message_headers_append (request_headers,
                                              "X-Nickname",
                                              nick);
-        soup_message_set_request (ret,
-                                  "application/json",
-                                  SOUP_MEMORY_TAKE,
-                                  data,
-                                  data_len);
+        body = g_bytes_new_take (data, data_len);
+        soup_message_set_request_body_from_bytes (ret, "application/json", body);
         g_debug ("Sending following request to '%s':\n%s", url, data);
 
+        mozilla->priv->bss_submitted = TRUE;
+        mozilla->priv->tower_submitted = TRUE;
+
 out:
         return ret;
 }
@@ -485,3 +582,134 @@ gclue_mozilla_should_ignore_bss (WPABSS *bss)
 
         return FALSE;
 }
+
+static void
+gclue_mozilla_finalize (GObject *object)
+{
+        GClueMozilla *mozilla = GCLUE_MOZILLA (object);
+
+        g_clear_weak_pointer (&mozilla->priv->wifi);
+
+        G_OBJECT_CLASS (gclue_mozilla_parent_class)->finalize (object);
+}
+
+static void
+gclue_mozilla_init (GClueMozilla *mozilla)
+{
+        mozilla->priv = gclue_mozilla_get_instance_private (mozilla);
+        mozilla->priv->wifi = NULL;
+        mozilla->priv->tower_valid = FALSE;
+        mozilla->priv->bss_submitted = FALSE;
+}
+
+static void
+gclue_mozilla_class_init (GClueMozillaClass *klass)
+{
+        GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
+        object_class->finalize = gclue_mozilla_finalize;
+}
+
+GClueMozilla *
+gclue_mozilla_get_singleton (void)
+{
+        static GClueMozilla *mozilla = NULL;
+
+        if (!mozilla) {
+                mozilla = g_object_new (GCLUE_TYPE_MOZILLA, NULL);
+                g_object_add_weak_pointer (G_OBJECT (mozilla), (gpointer) &mozilla);
+        } else
+                g_object_ref (mozilla);
+
+        return mozilla;
+}
+
+void
+gclue_mozilla_set_wifi (GClueMozilla *mozilla,
+                        GClueWifi *wifi)
+{
+        g_return_if_fail (GCLUE_IS_MOZILLA (mozilla));
+
+        if (mozilla->priv->wifi == wifi)
+                return;
+
+        g_clear_weak_pointer (&mozilla->priv->wifi);
+
+        if (!wifi) {
+                return;
+        }
+
+        mozilla->priv->wifi = wifi;
+        g_object_add_weak_pointer (G_OBJECT (mozilla->priv->wifi),
+                                   (gpointer) &mozilla->priv->wifi);
+}
+
+gboolean
+gclue_mozilla_test_set_wifi (GClueMozilla *mozilla,
+                             GClueWifi *old, GClueWifi *new)
+{
+        if (mozilla->priv->wifi != old)
+                return FALSE;
+
+        gclue_mozilla_set_wifi (mozilla, new);
+        return TRUE;
+}
+
+void
+gclue_mozilla_set_bss_dirty (GClueMozilla *mozilla)
+{
+        g_return_if_fail (GCLUE_IS_MOZILLA (mozilla));
+
+        mozilla->priv->bss_submitted = FALSE;
+}
+
+static gboolean gclue_mozilla_tower_identical (const GClue3GTower *t1,
+                                               const GClue3GTower *t2)
+{
+        return g_strcmp0 (t1->opc, t2->opc) == 0 && t1->lac == t2->lac &&
+                t1->cell_id == t2->cell_id && t1->tec == t2->tec;
+}
+
+void
+gclue_mozilla_set_tower (GClueMozilla *mozilla,
+                         const GClue3GTower *tower)
+{
+        g_return_if_fail (GCLUE_IS_MOZILLA (mozilla));
+
+        if (!tower ||
+            !(tower->tec > GCLUE_TOWER_TEC_UNKNOWN
+              && tower->tec <= GCLUE_TOWER_TEC_MAX_VALID)) {
+                mozilla->priv->tower_valid = FALSE;
+                return;
+        }
+
+        if (mozilla->priv->tower_valid &&
+            mozilla->priv->tower_submitted) {
+                mozilla->priv->tower_submitted =
+                        gclue_mozilla_tower_identical (&mozilla->priv->tower,
+                                                       tower);
+        } else
+                mozilla->priv->tower_submitted = FALSE;
+
+        mozilla->priv->tower = *tower;
+        mozilla->priv->tower_valid = TRUE;
+}
+
+gboolean
+gclue_mozilla_has_tower (GClueMozilla *mozilla)
+{
+        g_return_val_if_fail (GCLUE_IS_MOZILLA (mozilla), FALSE);
+
+        return mozilla->priv->tower_valid;
+}
+
+GClue3GTower *
+gclue_mozilla_get_tower (GClueMozilla *mozilla)
+{
+        g_return_val_if_fail (GCLUE_IS_MOZILLA (mozilla), NULL);
+
+        if (!mozilla->priv->tower_valid)
+                return NULL;
+
+        return &mozilla->priv->tower;
+}
